"""
These Task classes are "core" in the sense that they are used elsewhere in Prefect.
"""

import prefect.tasks.core.collections
import prefect.tasks.core.constants
import prefect.tasks.core.function
import prefect.tasks.core.operators
